@@ -41,17 +41,13 @@ import shlex
 import operator
 import subprocess
 import math
-# Setuptools ships with 'packaging' module, let's use it from there
-try:
-    from pkg_resources.extern.packaging.version import LegacyVersion
-# Fallback to a normal 'packaging' module in case 'setuptools' is stripped
-except:
-    from packaging.version import LegacyVersion
+import errno
 from socket import getfqdn
 from ldapurl import LDAPUrl
 from contextlib import closing
 
 import lib389
+from pathlib import Path
 from lib389.paths import ( Paths, DEFAULTS_PATH )
 from lib389.dseldif import DSEldif
 from lib389._constants import (
@@ -70,6 +66,8 @@ DEBUGGING = os.getenv('DEBUGGING', default=False)
 
 log = logging.getLogger(__name__)
 
+selinux_fcontext_info = None
+
 #
 # Various searches to be used in getEntry
 #   eg getEntry(*searches['NAMINGCONTEXTS'])
@@ -184,6 +182,7 @@ _chars = {
 # Utilities
 #
 
+
 def selinux_present():
     """
     Determine if selinux libraries are on a system, and if so, if we are in
@@ -200,6 +199,8 @@ def selinux_present():
             # (just checking the uid for now - we may rather check if semanage command success)
             if os.geteuid() != 0:
                 log.info('Non privileged user cannot use semanage, will not relabel ports or files.' )
+            elif not os.path.exists('/usr/sbin/semanage'):
+                log.info('semanage command is not installed, will not relabel ports or files. Please install policycoreutils-python-utils.' )
             else:
                 status = True
         else:
@@ -236,6 +237,120 @@ def selinux_restorecon(path):
         log.debug("Failed to run restorecon on: " + path)
 
 
+def _parse_semanage_fcontexts(cmd, regex=r"^(/[^ ]*)[^:=]+:[^:]*:([^:]*):.*$", reject=None):
+    """Parse semanage fcontext -L output
+    """
+    if reject is None:
+        reject = {}
+    info = {}
+    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    args = ' '.join(ensure_list_str(result.args))
+    stdout = ensure_str(result.stdout)
+    stderr = ensure_str(result.stderr)
+    if result.returncode:
+        log.debug("CMD: {args} returned {result.returncode} STDOUT: {stdout} STDERR: {stderr}")
+        return info
+    for m in re.finditer(regex, stdout, flags=re.MULTILINE):
+        if not m.group(1) in reject:
+            info[m.group(1)] = m.group(2)
+    return info
+
+
+def _get_selinux_fcontext_info():
+    global selinux_fcontext_info
+    if not selinux_fcontext_info:
+        local = _parse_semanage_fcontexts(["semanage", "fcontext", "-l", "-C"])
+        selinux_fcontext_info = {
+            "local"  : local,
+            "policy" : _parse_semanage_fcontexts(["semanage", "fcontext", "-l"], reject=local),
+            "equiv"  : _parse_semanage_fcontexts(["semanage", "fcontext", "-l"], regex=r"^(/[^ ]*) += +(.*)$"),
+        }
+    return selinux_fcontext_info
+
+
+def resolve_selinux_path(path):
+    '''Return the path as expected by semanage fcontext'''
+    path = str(Path(path).resolve())
+    if selinux_present():
+        _get_selinux_fcontext_info()
+        equiv = selinux_fcontext_info['equiv']
+        for r in equiv:
+            if path.startswith(r):
+                path = path.replace(r, equiv[r])
+    return path
+
+
+def selinux_label_file(path, label):
+    """
+    Set set an SELinux label to a file (or a directory)
+
+    :param path: The file path
+    :type path: str
+    :param label: The label to set (None to unset a label)
+    :type path: str
+    :raises: ValueError: Error message
+    """
+    if not selinux_present():
+        return
+    _get_selinux_fcontext_info()
+    if os.path.exists(path):
+        path = resolve_selinux_path(path)
+    local = selinux_fcontext_info['local']
+    policy = selinux_fcontext_info['policy']
+    if path in local:
+        if local[path] == label:
+            return
+        log.debug(f"Removing SELinux file context {path} with label {local[path]}.")
+        subprocess.run(["semanage", "fcontext", "-d", path])
+        del local[path]
+    if path in policy:
+        if policy[path] == label:
+            return
+        raise ValueError(f'Cannot change file context for {path} because it is defined in SELinux policy. Please choose another path.')
+    if label:
+        try:
+            log.debug(f"Setting label {label} in SELinux file context {path}.")
+            result = subprocess.run(["semanage", "fcontext", "-a", "-t", label, path],
+                                     stdout=subprocess.PIPE,
+                                     stderr=subprocess.PIPE)
+            args = ' '.join(ensure_list_str(result.args))
+            stdout = ensure_str(result.stdout)
+            stderr = ensure_str(result.stderr)
+            if result.returncode != 0:
+                log.error(f"ERROR CMD: {args} ; STDOUT: {stdout} ; STDERR: {stderr}")
+                result.check_returncode()
+            local[path] = label
+        except (OSError, subprocess.CalledProcessError) as e:
+            raise ValueError(f"Failed to set SElinux label {label} on {path}: {str(e)}")
+    if os.path.exists(path):
+        # pytest fails if I use selinux_restorecon(path)
+        subprocess.run(["restorecon", "-R", path])
+
+
+def selinux_clean_files_label(all=False):
+    """Remove dirsrv_*_t labels from either removed files or all files."""
+    if not selinux_present():
+        return
+    _get_selinux_fcontext_info()
+    local = { **selinux_fcontext_info['local'] }
+    for path, label in local.items():
+        if label in ( 'dirsrv_config_t', 'dirsrv_tmpfs_t', 'dirsrv_var_lib_t', 'dirsrv_var_lock_t', 'dirsrv_var_log_t', 'dirsrv_var_run_t', ):
+            if all or not os.path.exists(path):
+                selinux_label_file(path, None)
+
+
+def selinux_clean_ports_label():
+    """Remove labels from all port having ldap_port_t labels."""
+    if not selinux_present():
+        return
+    cmd=[ 'semanage', 'port', '--list', '-C' ]
+    res = _parse_semanage_fcontexts(cmd, regex=r"^(ldap_port_t) *tcp *(.*)$")
+    if 'ldap_port_t' in res:
+        for port in res['ldap_port_t'].split(","):
+            # Cannot use selinux_label_port here because port may be a range.
+            subprocess.run(["semanage", "port", "-d", "-p", "tcp", port.strip()])
+
+
 def _get_selinux_port_policies(port):
     """Get a list of selinux port policies for the specified port, 'tcp' protocol and
     excluding 'unreserved_port_t', 'reserved_port_t', 'ephemeral_port_t' labels"""
@@ -1094,6 +1209,76 @@ def generate_ds_params(inst_num, role=ReplicaRole.STANDALONE):
 
     return instance_data
 
+class DSVersion():
+    def __init__(self, version):
+        self._version = str(version)
+        self._key = _cmpkey(self._version)
+
+    def __str__(self):
+        return self._version
+
+    def __repr__(self):
+        return f"<DSVersion('{self}')>"
+
+    def __hash__(self):
+        return hash(self._key)
+
+    def __lt__(self, other):
+        if not isinstance(other, DSVersion):
+            return NotImplemented
+
+        return self._key < other._key
+
+    def __le__(self, other):
+        if not isinstance(other, DSVersion):
+            return NotImplemented
+
+        return self._key <= other._key
+
+    def __eq__(self, other):
+        if not isinstance(other, DSVersion):
+            return NotImplemented
+
+        return self._key == other._key
+
+    def __ge__(self, other):
+        if not isinstance(other, DSVersion):
+            return NotImplemented
+
+        return self._key >= other._key
+
+    def __gt__(self, other):
+        if not isinstance(other, DSVersion):
+            return NotImplemented
+
+        return self._key > other._key
+
+    def __ne__(self, other):
+        if not isinstance(other, DSVersion):
+            return NotImplemented
+
+        return self._key != other._key
+
+
+def _parse_version_parts(s):
+    for part in re.compile(r"(\d+ | [a-z]+ | \. | -)", re.VERBOSE).split(s):
+
+        if not part or part == ".":
+            continue
+
+        if part[:1] in "0123456789":
+            # pad for numeric comparison
+            yield part.zfill(8)
+        else:
+            yield "*" + part
+
+def _cmpkey(version):
+    parts = []
+    for part in _parse_version_parts(version.lower()):
+        parts.append(part)
+
+    return tuple(parts)
+
 
 def get_ds_version(paths=None):
     """
@@ -1121,9 +1306,9 @@ def ds_is_related(relation, *ver, instance=None):
     if len(ver) > 1:
         for cmp_ver in ver:
             if cmp_ver.startswith(ds_ver[:3]):
-                return ops[relation](LegacyVersion(ds_ver),LegacyVersion(cmp_ver))
+                return ops[relation](DSVersion(ds_ver),DSVersion(cmp_ver))
     else:
-        return ops[relation](LegacyVersion(ds_ver), LegacyVersion(ver[0]))
+        return ops[relation](DSVersion(ds_ver), DSVersion(ver[0]))
 
 
 def ds_is_older(*ver, instance=None):
@@ -1209,6 +1394,32 @@ def socket_check_open(host, port):
                 return False
 
 
+def socket_check_bind(port):
+    """
+    Check if a socket can be bound.  Need to handle cases where IPv6 is
+    completely disabled.
+    """
+    # Trying IPv6 first ...
+    host = "::1"
+    family = socket.AF_INET6
+    try:
+        with closing(socket.socket(family, socket.SOCK_STREAM)) as sock:
+            sock.bind((host, port))
+            return True
+    except OSError as e:
+        if e.errno == errno.EACCES or  e.errno == errno.EADDRINUSE:
+            return False
+    # Maybe there is no IPv6, adjust hostname, and try IPv4 ...
+    host = "127.0.0.1"
+    family = socket.AF_INET
+    try:
+        with closing(socket.socket(family, socket.SOCK_STREAM)) as sock:
+            sock.bind((host, port))
+            return True
+    except OSError as e:
+        return False
+
+
 def ensure_bytes(val):
     if val is not None and type(val) != bytes:
         return val.encode()
@@ -1326,12 +1537,11 @@ def get_instance_list():
                 insts.append(inst)
     except OSError as e:
         log.error("Failed to check directory: {} - {}".format(conf_dir, str(e)))
-    except IOError as e:
-        log.error(e)
-        log.error("Perhaps you need to be a different user?")
+
     insts.sort()
     return insts
 
+
 # used by arg parse completers
 def instance_choices(prefix, parsed_args, **kwargs):
     insts = get_instance_list()
@@ -1342,6 +1552,7 @@ def instance_choices(prefix, parsed_args, **kwargs):
 
     return inst_names
 
+
 def get_user_is_ds_owner():
     # Check if we have permission to administer the DS instance. This is required
     # for some tasks such as installing, killing, or editing configs for the
@@ -1469,6 +1680,7 @@ def is_valid_hostname(hostname):
     allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
     return all(allowed.match(x) for x in hostname.split("."))
 
+
 def get_default_db_lib():
     return os.getenv('NSSLAPD_DB_LIB', default=DEFAULT_DB_LIB)
 
@@ -1517,6 +1729,7 @@ def elapsed_time(timestamp1, timestamp2):
     else:
         return ""
 
+
 def get_task_status(inst, log, taskObj, dn=None, show_log=False, watch=False, use_json=False):
     """Get the status of a Task, if "watching" keep looping over the tasks
     :param inst: DirSrv Object
@@ -1561,15 +1774,7 @@ def get_task_status(inst, log, taskObj, dn=None, show_log=False, watch=False, us
                 all_finished = False
                 task_ended = ""
                 exitcode = "Not finished ..."
-
-                # Calc elapsed time for running task
-                curr_time = int(time.time())  # Use the current time
-                time_tuple = (int(task_created[:4]), int(task_created[4:6]), int(task_created[6:8]),
-                              int(task_created[8:10]), int(task_created[10:12]), int(task_created[12:14]),
-                              0, 0, 0)
-                start_time = int(time.mktime(time_tuple))
-                elapsed_secs = curr_time - start_time
-                elapsed_time_str = str(timedelta(seconds=elapsed_secs))
+                elapsed_time_str = ""
             else:
                 # Task is finished, use start and end times to calc elapsed time
                 elapsed_time_str = elapsed_time(task_created, task_ended)
